@@ -15,13 +15,14 @@
 #include <sys/sysmacros.h>
 #include <sys/poll.h>
 #include <linux/major.h>
+#include <sys/time.h>
 typedef unsigned char u_char;   /* horrible, for scsi.h */
 #include "sg_include.h"
 #include "sg_err.h"
 #include "llseek.h"
 
 /* A utility program for the Linux OS SCSI generic ("sg") device driver.
-*  Copyright (C) 1999, 2000 D. Gilbert and P. Allworth
+*  Copyright (C) 1999-2002 D. Gilbert and P. Allworth
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
@@ -47,7 +48,7 @@ typedef unsigned char u_char;   /* horrible, for scsi.h */
 
 */
 
-static char * version_str = "0.52 20010819";
+static char * version_str = "0.55 20020509";
 
 #define DEF_BLOCK_SIZE 512
 #define DEF_BLOCKS_PER_TRANSFER 128
@@ -79,6 +80,11 @@ static char * version_str = "0.52 20010819";
 #define QS_IN_POLL 11
 #define QS_OUT_POLL 12
 
+#define STR_SZ 1024
+#define INOUTF_SZ 512
+#define EBUFF_SZ 512
+
+
 struct request_element;
 
 typedef struct request_collection
@@ -209,16 +215,18 @@ int dd_filetype(const char * filename)
 void usage()
 {
     fprintf(stderr, "Usage: "
-           "sgq_dd  [if=<infile>] [skip=<n>] [of=<ofile>] [seek=<n>]\n"
-           "               [bs=<num>] [bpt=<num>] [count=<n>]\n"
-           "               [dio=<n>] [thr=<n>] [coe=<n>] [gen=<n>]\n"
-           "               [deb=<n>] [--version]\n"
-           "            usually either 'if' or 'of' is a sg or raw device\n"
+           "sgq_dd  [if=<infile>] [skip=<n>] [of=<ofile>] [seek=<n>] "
+           "[bs=<num>]\n"
+           "            [bpt=<num>] [count=<n>] [dio=0|1] [thr=<n>] "
+	   "[coe=0|1] [gen=<n>]\n"
+           "            [time=0|1] [deb=<n>] [--version]\n"
+           "         usually either 'if' or 'of' is a sg or raw device\n"
            " 'bpt' is blocks_per_transfer (default is 128)\n"
            " 'dio' is direct IO, 1->attempt, 0->indirect IO (def)\n"
            " 'thr' is number of queues, must be > 0, default 4, max 32\n");
     fprintf(stderr, " 'coe' continue on sg error, 0->exit (def), "
     	   "1->zero + continue\n"
+	   " 'time' 0->no timing(def), 1->time plus calculate throughput\n"
            " 'gen' 0-> 1 file is special(def), 1-> any files allowed\n"
            " 'deb' is debug, 0->none (def), > 0->varying degrees of debug\n");
 }
@@ -467,8 +475,8 @@ int sg_start_io(Rq_elem * rep)
         fprintf(stderr, "sg_start_io: SCSI %s, blk=%d num_blks=%d\n",
                rep->wr ? "WRITE" : "READ", rep->blk, rep->num_blks);
         sg_print_command(hp->cmdp);
-        fprintf(stderr, "dir=%d, len=%d, dxfrp=%p, cmd_len=%d\n",
-                hp->dxfer_direction, hp->dxfer_len, hp->dxferp, hp->cmd_len);
+        fprintf(stderr, " len=%d, dxfrp=%p, cmd_len=%d\n",
+                hp->dxfer_len, hp->dxferp, hp->cmd_len);
     }
 
     while (((res = write(rep->wr ? rep->outfd : rep->infd, hp,
@@ -524,9 +532,9 @@ int sg_finish_io(int wr, Rq_elem * rep)
             return 1;
         default:
             {
-                char ebuff[64];
-                sprintf(ebuff, "%s blk=%d", rep->wr ? "writing": "reading",
-                        rep->blk);
+                char ebuff[EBUFF_SZ];
+                snprintf(ebuff, EBUFF_SZ, "%s blk=%d", 
+			 rep->wr ? "writing": "reading", rep->blk);
                 sg_chk_n_print3(ebuff, hp);
                 return -1;
             }
@@ -541,7 +549,8 @@ int sg_finish_io(int wr, Rq_elem * rep)
         rep->dio_incomplete = 0;
     rep->resid = hp->resid;
     if (rep->debug > 8)
-        fprintf(stderr, "sg_finish_io: completed %s\n", wr ? "WRITE" : "READ");
+        fprintf(stderr, "sg_finish_io: completed %s, blk=%d\n", 
+		wr ? "WRITE" : "READ", rep->blk);
     return 0;
 }
 
@@ -580,7 +589,7 @@ int prepare_rq_elems(Rq_coll * clp, const char * inf, const char * outf)
     int k;
     Rq_elem * rep;
     size_t psz;
-    char ebuff[256];
+    char ebuff[EBUFF_SZ];
     int sz = clp->bpt * clp->bs;
     int scsi_type;
 
@@ -605,8 +614,8 @@ int prepare_rq_elems(Rq_coll * clp, const char * inf, const char * outf)
 		rep->infd = clp->infd;
 	    else {
 		if ((rep->infd = open(inf, O_RDWR)) < 0) {
-                    sprintf(ebuff, "sgq_dd: could not open %s for sg reading",
-                            inf);
+                    snprintf(ebuff, EBUFF_SZ, 
+			     "sgq_dd: could not open %s for sg reading", inf);
                     perror(ebuff);
                     return 1;
                 }
@@ -627,8 +636,8 @@ int prepare_rq_elems(Rq_coll * clp, const char * inf, const char * outf)
 		rep->outfd = clp->outfd;
 	    else {
 		if ((rep->outfd = open(outf, O_RDWR)) < 0) {
-                    sprintf(ebuff, "sgq_dd: could not open %s for sg writing",
-                            outf);
+                    snprintf(ebuff, EBUFF_SZ, 
+			     "sgq_dd: could not open %s for sg writing", outf);
                     perror(ebuff);
                     return 1;
                 }
@@ -683,6 +692,13 @@ int decider(Rq_coll * clp, int first_xfer, int * req_indexp)
 	    *req_indexp = lowest_blk_index;
 	return QS_IN_FINISHED;
     }
+#if 0
+    if (try_poll) {
+	res = do_poll(clp, 2, req_indexp);
+	if (0 != res)
+	    return res;
+    }
+#endif
     if (req_indexp)
     	*req_indexp = first_idle_index;
     return QS_IDLE;
@@ -732,20 +748,22 @@ int main(int argc, char * argv[])
     int seek = 0;
     int ibs = 0;
     int obs = 0;
-    char str[512];
+    char str[STR_SZ];
     char * key;
     char * buf;
-    char inf[512];
-    char outf[512];
+    char inf[INOUTF_SZ];
+    char outf[INOUTF_SZ];
     int res, k;
     int in_num_sect = 0;
     int out_num_sect = 0;
     int num_threads = DEF_NUM_THREADS;
     int gen = 0;
+    int do_time = 0;
     int in_sect_sz, out_sect_sz, first_xfer, qstate, req_index, seek_skip;
     int blocks, stop_after_write, terminate;
-    char ebuff[256];
+    char ebuff[EBUFF_SZ];
     Rq_elem * rep;
+    struct timeval start_tm, end_tm;
 
     memset(&rcoll, 0, sizeof(Rq_coll));
     rcoll.bpt = DEF_BLOCKS_PER_TRANSFER;
@@ -760,7 +778,7 @@ int main(int argc, char * argv[])
 
     for(k = 1; k < argc; k++) {
         if (argv[k])
-            strcpy(str, argv[k]);
+            strncpy(str, argv[k], STR_SZ);
         else
             continue;
         for(key = str, buf = key; *buf && *buf != '=';)
@@ -768,9 +786,9 @@ int main(int argc, char * argv[])
         if (*buf)
             *buf++ = '\0';
         if (strcmp(key,"if") == 0)
-            strcpy(inf, buf);
+            strncpy(inf, buf, INOUTF_SZ);
         else if (strcmp(key,"of") == 0)
-            strcpy(outf, buf);
+            strncpy(outf, buf, INOUTF_SZ);
         else if (0 == strcmp(key,"ibs"))
             ibs = get_num(buf);
         else if (0 == strcmp(key,"obs"))
@@ -795,6 +813,8 @@ int main(int argc, char * argv[])
             gen = get_num(buf);
         else if (0 == strncmp(key,"deb", 3))
             rcoll.debug = get_num(buf);
+        else if (0 == strcmp(key,"time"))
+            do_time = get_num(buf);
         else if (0 == strncmp(key, "--vers", 6)) {
             fprintf(stderr, "sgq_dd for sg version 3 driver: %s\n", 
 	    	    version_str);
@@ -840,15 +860,16 @@ int main(int argc, char * argv[])
 
         if (FT_SG == rcoll.in_type) {
             if ((rcoll.infd = open(inf, O_RDWR)) < 0) {
-                sprintf(ebuff, "sgq_dd: could not open %s for sg reading", 
-			inf);
+                snprintf(ebuff, EBUFF_SZ,
+			 "sgq_dd: could not open %s for sg reading", inf);
                 perror(ebuff);
                 return 1;
             }
         }
         if (FT_SG != rcoll.in_type) {
             if ((rcoll.infd = open(inf, O_RDONLY)) < 0) {
-                sprintf(ebuff, "sgq_dd: could not open %s for reading", inf);
+                snprintf(ebuff, EBUFF_SZ,
+			 "sgq_dd: could not open %s for reading", inf);
                 perror(ebuff);
                 return 1;
             }
@@ -857,7 +878,7 @@ int main(int argc, char * argv[])
 
                 offset *= rcoll.bs;       /* could exceed 32 here! */
                 if (llse_llseek(rcoll.infd, offset, SEEK_SET) < 0) {
-                    sprintf(ebuff,
+                    snprintf(ebuff, EBUFF_SZ,
                 "sgq_dd: couldn't skip to required position on %s", inf);
                     perror(ebuff);
                     return 1;
@@ -870,7 +891,7 @@ int main(int argc, char * argv[])
 
         if (FT_SG == rcoll.out_type) {
 	    if ((rcoll.outfd = open(outf, O_RDWR)) < 0) {
-                sprintf(ebuff, 
+                snprintf(ebuff, EBUFF_SZ,
 			"sgq_dd: could not open %s for sg writing", outf);
                 perror(ebuff);
                 return 1;
@@ -879,7 +900,7 @@ int main(int argc, char * argv[])
 	else {
 	    if (FT_OTHER == rcoll.out_type) {
 		if ((rcoll.outfd = open(outf, O_WRONLY | O_CREAT, 0666)) < 0) {
-                    sprintf(ebuff,
+                    snprintf(ebuff, EBUFF_SZ,
                             "sgq_dd: could not open %s for writing", outf);
                     perror(ebuff);
                     return 1;
@@ -887,7 +908,7 @@ int main(int argc, char * argv[])
 	    }
 	    else {
 		if ((rcoll.outfd = open(outf, O_WRONLY)) < 0) {
-                    sprintf(ebuff,
+                    snprintf(ebuff, EBUFF_SZ,
                             "sgq_dd: could not open %s for raw writing", outf);
                     perror(ebuff);
                     return 1;
@@ -898,7 +919,7 @@ int main(int argc, char * argv[])
 
                 offset *= rcoll.bs;       /* could exceed 32 bits here! */
 		if (llse_llseek(rcoll.outfd, offset, SEEK_SET) < 0) {
-                    sprintf(ebuff,
+                    snprintf(ebuff, EBUFF_SZ,
                 "sgq_dd: couldn't seek to required position on %s", outf);
                     perror(ebuff);
                     return 1;
@@ -987,6 +1008,11 @@ int main(int argc, char * argv[])
     stop_after_write = 0;
     terminate = 0;
     seek_skip =  rcoll.seek - rcoll.skip;
+    if (do_time) {
+        start_tm.tv_sec = 0;
+        start_tm.tv_usec = 0;
+        gettimeofday(&start_tm, NULL);
+    }
     while (rcoll.out_done_count > 0) { /* >>>>>>>>> main loop */
         req_index = -1;
 	qstate = decider(&rcoll, first_xfer, &req_index);
@@ -995,7 +1021,8 @@ int main(int argc, char * argv[])
 	case QS_IDLE:
 	    if ((NULL == rep) || (rcoll.in_count <= 0)) {
 	    	/* usleep(1000); */
-		do_poll(&rcoll, 10, NULL);
+		/* do_poll(&rcoll, 10, NULL); */
+		/* do_poll(&rcoll, 0, NULL); */
 		break;
 	    }
 	    if (rcoll.debug > 8)
@@ -1038,7 +1065,10 @@ int main(int argc, char * argv[])
 				"req_index=%d\n", req_index);
 	    if ((rep->blk + seek_skip) != rcoll.out_blk) {
 		/* if write would be out of sequence then wait */
-		usleep(1000);
+		if (rcoll.debug > 4)
+		    fprintf(stderr, "    sgq_dd: QS_IN_FINISHED, "
+			    "out of sequence\n");
+		usleep(200);
 		break;
 	    }
             rep->wr = 1;
@@ -1104,6 +1134,28 @@ int main(int argc, char * argv[])
 	    break;
     } /* >>>>>>>>>>>>> end of main loop */
 
+    if ((do_time) && (start_tm.tv_sec || start_tm.tv_usec)) {
+        struct timeval res_tm;
+        double a, b;
+
+        gettimeofday(&end_tm, NULL);
+        res_tm.tv_sec = end_tm.tv_sec - start_tm.tv_sec;
+        res_tm.tv_usec = end_tm.tv_usec - start_tm.tv_usec;
+        if (res_tm.tv_usec < 0) {
+            --res_tm.tv_sec;
+            res_tm.tv_usec += 1000000;
+        }
+        a = res_tm.tv_sec;
+        a += (0.000001 * res_tm.tv_usec);
+        b = (double)rcoll.bs * (dd_count - rcoll.out_done_count);
+        printf("time to transfer data was %d.%06d secs",
+               (int)res_tm.tv_sec, (int)res_tm.tv_usec);
+        if ((a > 0.00001) && (b > 511))
+            printf(", %.2f MB/sec\n", b / (a * 1000000.0));
+        else
+            printf("\n");
+    }
+
     if (STDIN_FILENO != rcoll.infd)
         close(rcoll.infd);
     if (STDOUT_FILENO != rcoll.outfd)
